#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <bios.h>
#include <dos.h>
#include <conio.h>


#define BASE_PORT1	0x2a20
#define BASE_PORT2	0x2a30


#define _BV(x)		(1 << (x))

#define PIC_REG_EOI	0x20

#define REG_PIU0	0x0
#define REG_PIU1	0x1
#define REG_PIU2	0x2
#define REG_PCR		0x3
#define REG_CNTR0	0x4
#define REG_CNTR1	0x5
#define REG_CNTR2	0x6
#define REG_TCWR	0x7
#define REG_TCR		0x8
#define REG_TSR		0xc

#define PIU2_EXR8	_BV(7)
#define PIU2_RIE	_BV(4)
#define PIU2_RXRDY	_BV(5)
//#define PIU2_RXRDY	_BV(3)
#define PIU2_WIE	_BV(2)
#define PIU2_TXRDY	_BV(0)

#define TCR_IBE		_BV(7)
#define TCR_TMSK	_BV(6)
#define TCR_EXT8	_BV(4)
#define TCR_TBE		_BV(3)
#define TCR_TAE		_BV(2)
#define TCR_TBC		_BV(1)
#define TCR_TAC		_BV(0)

#define TSR_TCS		_BV(7)
#define TSR_TBS		_BV(1)
#define TSR_TAS		_BV(0)


static int tcr = 0;

static unsigned int far *sys_clock = (unsigned int far *)0x0000046c;


// forward declarations
void tcr_set(unsigned port_base, int val);
void tcr_clear(unsigned port_base, int val);


// ...
int get_data(unsigned port_base)
{
	unsigned char stat;
	unsigned int time, time_start;
	int data;

	time_start = *sys_clock;
	while(((stat = inp(port_base + REG_PIU2)) & PIU2_RXRDY) == 0) {
		time = *sys_clock - time_start;
		if (time > 20) {
			return -1;
		}
	}

	data = inp(port_base + REG_PIU0);
	if (stat & PIU2_EXR8) data |= 0x0100;
	return data;
}


int put_data(unsigned port_base, int data)
{
	unsigned char stat;
	unsigned int time, time_start;

	time_start = *sys_clock;
	while(((stat = inp(port_base + REG_PIU2)) & PIU2_TXRDY) == 0) {
		time = *sys_clock - time_start;
		if (time > 20) {
//			printf("put_data() failed: stat=0x%02x\r\n", stat);
			return -1;
		}
	}

	if (data & 0x0100) {
		tcr_set(port_base, TCR_EXT8);
	} else {
		tcr_clear(port_base, TCR_EXT8);
	}
	outp(port_base + REG_PIU1, data & 0xff);

//	printf("put_data() ok: 0x%03x\r\n", data);
	return 0;
}


int get_tsr(unsigned port_base)
{
	return inp(port_base + REG_TSR);
}

void tcr_reset(unsigned port_base)
{
	tcr = 0;
	outp(port_base + REG_TCR, tcr);
}

void tcr_set(unsigned port_base, int val)
{
	tcr |= val;
	outp(port_base + REG_TCR, tcr);
//	printf("TCR=0x%x\r\n", tcr);
}

void tcr_clear(unsigned port_base, int val)
{
	tcr &= ~val;
	outp(port_base + REG_TCR, tcr);
//	printf("TCR=0x%x\r\n", tcr);
}

void tcr_rearm(unsigned port_base, int val)
{
	tcr_clear(port_base, val);
	tcr_set(port_base, val);
}


// timer
unsigned timer_get(unsigned port_base, int timer)
{
	unsigned val;
	switch(timer) {
	case 0:
		// latches contents of counter 0 to the register
		outp(port_base + REG_TCWR, 0x00);
		val = inp(port_base + REG_CNTR0);
		val = val | (inp(port_base + REG_CNTR0) << 8);
		break;
	case 1:
		// latches contents of counter 1 to the register
		outp(port_base + REG_TCWR, 0x40);
		val = inp(port_base + REG_CNTR1);
//		val = val | (inp(port_base + REG_CNTR1) << 8);
		break;
	case 2:
		// latches contents of counter 2 to the register
		outp(port_base + REG_TCWR, 0x80);
		val = inp(port_base + REG_CNTR2);
//		val = val | (inp(port_base + REG_CNTR2) << 8);
		break;
	default:
		val = 0;
	}

	return val;
}

void timer_set(unsigned port_base, int timer, unsigned val)
{
	switch(timer) {
	case 0:
		// latches contents of counter 0 to the register
		outp(port_base + REG_TCWR, 0x34);
		outp(port_base + REG_CNTR0, val & 0xff);
		outp(port_base + REG_CNTR0, val >> 8);
		break;
	case 1:
		// latches contents of counter 1 to the register
		outp(port_base + REG_TCWR, 0x74);
		outp(port_base + REG_CNTR1, val & 0xff);
		outp(port_base + REG_CNTR1, val >> 8);
		break;
	case 2:
		// latches contents of counter 2 to the register
		outp(port_base + REG_TCWR, 0xb6);
		outp(port_base + REG_CNTR2, val & 0xff);
		outp(port_base + REG_CNTR2, val >> 8);
		break;
	}
}

void timer_b_poll(void)
{
	int tsr;
	int i;

	tcr_set(BASE_PORT1, TCR_TBE | TCR_TBC);

	timer_set(BASE_PORT1, 1, 200);
	timer_set(BASE_PORT1, 2, 10000);

	for (i = 0; i < 160; i++) {
		tsr = get_tsr(BASE_PORT1);
		if (tsr & TSR_TBS) {
			tcr_rearm(BASE_PORT1, TCR_TBC);
			printf("Poll B\r\n");
		}
		delay(100);
	}
}


// timer interrupt test

#define IMFC_INT	0x0d

static unsigned int_cnt = 0;
void interrupt (*dos_isr)();

void interrupt imfc_isr()
{
	int tsr;

	_disable();

	// mask IMFC interrupt
	tcr_clear(BASE_PORT1, TCR_TMSK);


	// accept interrupt
	outp(PIC_REG_EOI, 0x20);

	_enable();

	tsr = get_tsr(BASE_PORT1);
	if (tsr & TSR_TBS) {
		int_cnt++;
		tcr_rearm(BASE_PORT1, TCR_TBC);
	}

	tcr_set(BASE_PORT1, TCR_TMSK);
}


void timer_b_int()
{
	int tsr;
	int i;
	unsigned old_int_cnt;

	printf("Timer B interrupt test\r\n");

	tsr = inp(0x21);
	printf("MSK=0x%x\r\n", tsr);
	outp(0x21, tsr & ~(1 << 5));

	// IRQ5 = INT_D
	_disable();

	dos_isr = _dos_getvect(IMFC_INT);
	_dos_setvect(IMFC_INT, imfc_isr);

	timer_set(BASE_PORT1, 1, 200);
	timer_set(BASE_PORT1, 2, 10000);

	tcr_set(BASE_PORT1, TCR_TBE | TCR_TBC);
	tcr_set(BASE_PORT1, TCR_TMSK | TCR_IBE);

	// unmask PIC1

	_enable();

	old_int_cnt = int_cnt;
	for (i = 0; i < 160; i++) {
//		tsr = get_tsr(BASE_PORT1);
//		printf("Int B counter %u, TSR=0x%x\r\n", int_cnt, tsr);
		if (int_cnt != old_int_cnt) {
			old_int_cnt = int_cnt;
			printf("Int B\r\n");
		}
		delay(100);
	}

	_disable();

	tsr = inp(0x21);
	printf("MSK=0x%x\r\n", tsr);
	outp(0x21, tsr | (1 << 5));

	tcr_clear(BASE_PORT1, TCR_TMSK | TCR_IBE);
	_dos_setvect(IMFC_INT, dos_isr);

	_enable();
}



int imfc_init(unsigned port_base) {
	unsigned char stat;

	stat = inp(port_base + REG_PIU2);
	if (stat == 0xff) {
		printf("IMFC not present\r\n");
		return -1;
	}

	// initialize PIU
	outp(port_base + REG_PCR, 0xbc);
	outp(port_base + REG_PCR, 0x05);

	tcr_reset(port_base);

	printf("IMFC init Ok\r\n");

	return 0;
}


int imfc_set_path(unsigned port_base, int p1, int p2, int p3, int p4, int p5)
{
	int ret;

	ret = put_data(port_base, 0x1e2);
	if (ret < 0) return ret;

	ret = put_data(port_base, 0x100 | p1);
	if (ret < 0) return ret;

	ret = put_data(port_base, 0x100 | p2);
	if (ret < 0) return ret;

	ret = put_data(port_base, 0x100 | p3);
	if (ret < 0) return ret;

	ret = put_data(port_base, 0x100 | p4);
	if (ret < 0) return ret;

	ret = put_data(port_base, 0x100 | p5);
	if (ret < 0) return ret;

//	printf("set path Ok\r\n");
	return 0;
}


int imfc_note(unsigned port_base, int chan, int pitch, int velocity)
{
	(void)put_data(port_base, 0x90 | ((chan - 1) & 0x0f));
	(void)put_data(port_base, pitch & 0x7f);
	(void)put_data(port_base, velocity & 0x7f);
}


#define TEST_SYNTH_NOTES_NUM	9
static int test_synth_notes[TEST_SYNTH_NOTES_NUM] = {
	60, 62, 64, 65, 67, 65, 64, 62, 60
};

int test_synth(unsigned port_base)
{
	int ret, i;

//	ret = imfc_set_path(BASE_PORT1, 0x00, 0x00, 0x00, 0x1f, 0x00);
	ret = imfc_set_path(BASE_PORT1, 0x00, 0x1f, 0x1f, 0x00, 0x00);
	if (ret < 0) return ret;

	for (i = 0; i < TEST_SYNTH_NOTES_NUM; i++) {
		imfc_note(BASE_PORT1, 1, test_synth_notes[i], 64);
		delay(500);
		imfc_note(BASE_PORT1, 1, test_synth_notes[i], 0);
	}

	return 0;
}


#define TEST_LOOP_DATA_NUM	37
static int test_loop_data[TEST_LOOP_DATA_NUM] = {
	0x90,
	60, 64,	60, 0,
	62, 64,	62, 0,
	64, 64,	64, 0,
	65, 64,	65, 0,
	67, 64,	67, 0,
	65, 64,	65, 0,
	64, 64,	64, 0,
	62, 64,	62, 0,
	60, 64,	60, 0
};

int test_loop(unsigned port_base)
{
	int data;
	int ret, i;

	ret = imfc_set_path(BASE_PORT1, 0x1f, 0x1f, 0x00, 0x00, 0x00);
	if (ret < 0) return ret;

	// flush input FIFO
	do {
		data = get_data(BASE_PORT1);
	} while(data >= 0);

	for (i = 0; i < TEST_LOOP_DATA_NUM; i++) {
		(void)put_data(port_base, test_loop_data[i]);
	}

	for (i = 0; i < TEST_LOOP_DATA_NUM; i++) {
		data = get_data(BASE_PORT1);
//		printf("[%02d]: 0x%03x - 0x%03x\r\n", i, test_loop_data[i], data);
		if (data != test_loop_data[i]) {
			printf("Loop test Failed\r\n");
			return -1;
		}
	}

	printf("Loop test Ok\r\n");
	return 0;
}



// entry point
int main(int argc, char *argv[])
{
	int i;
	unsigned c;
	int tsr;
	int data;
	int ret;

	ret = imfc_init(BASE_PORT1);
	if (ret < 0) return ret;

	ret = imfc_set_path(BASE_PORT1, 0x1f, 0x1f, 0x1f, 0x1f, 0x1f);
	if (ret < 0) return ret;

	// flush input FIFO
	do {
		data = get_data(BASE_PORT1);
	} while(data >= 0);

//	timer_set(BASE_PORT1, 0, 32);

//	timer_set(BASE_PORT1, 2, 10000);
//	timer_set(BASE_PORT1, 1, 200);

//	timer_b_poll();
//	timer_b_int();


//	test_synth(BASE_PORT1);
	test_loop(BASE_PORT1);

	return 0;
}
